@@ -6,6 +6,8 @@ from .transform import rotations
 
 
 class Shape(object):
+    bit_masks = None
+
     def is_contained(self, tile):
         return all(self.is_in(sq) for sq in tile)
 
@@ -25,17 +27,32 @@ class Shape(object):
                     yield translated
 
     def bit_vector(self, tile):
-        return [sq in tile for sq in self.squares]
+        return np.sum([self.bit_mask(sq) for sq in tile], axis=0)
+
+    def bit_mask(self, sq):
+        if not self.bit_masks:
+            self.bit_masks = dict()
+        if sq not in self.bit_masks:
+            self.bit_masks[sq] = np.array(
+                [s == sq for s in self.squares], dtype=np.int32
+            )
+        return self.bit_masks[sq]
 
     def remove(self, square):
         if square not in self.squares:
-            raise Exception("Tried to remove a square %s which was not present in the board" % (square, ))
+            raise Exception(
+                "Tried to remove a square %s which was not present in the board"
+                % (square,)
+            )
         return Irregular([sq for sq in self.squares if sq != square])
 
     def remove_all(self, squares):
         for square in squares:
             if square not in self.squares:
-                raise Exception("Tried to remove a square %s which was not present in the board" % (square, ))
+                raise Exception(
+                    "Tried to remove a square %s which was not present in the board"
+                    % (square,)
+                )
         return Irregular([sq for sq in self.squares if sq not in squares])
 
     def tile_with(self, tiles):
@@ -66,64 +83,12 @@ class Irregular(Shape):
     @property
     def adjusted(self):
         if self._adjusted is None:
-            self._adjusted = [
-                (x - self.min_x, y - self.min_y) for x, y in self.squares
-            ]
+            self._adjusted = [(x - self.min_x, y - self.min_y) for x, y in self.squares]
         return self._adjusted
 
     def is_in(self, square):
         return square in self._squares
 
-    def format_row_sides(self, row):
-        return " ".join("|" if r else " " for r in row)
-
-    def format_row_upper(self, row):
-        return "+" + "+".join("-" if r else " " for r in row) + "+"
-
-    def format_tiling_lines(self, h, v):
-        for i in range(0, self.max_y - self.min_y + 1):
-            yield self.format_row_upper(h[i])
-            yield self.format_row_sides(v[i])
-        yield self.format_row_upper(h[self.max_y - self.min_y + 1])
-
-    def make_base_h_row(self, i):
-        lines_above = set(x for x, y in self.squares if y == i - 1)
-        lines_below = set(x for x, y in self.squares if y == i)
-        lines = lines_above.union(lines_below)
-
-        return [(x in lines) for x in range(self.min_x, self.max_x + 1)]
-
-    def make_base_v_row(self, i):
-        lines_left = set(x for x, y in self.squares if y == i)
-        lines_right = set(x + 1 for x, y in self.squares if y == i)
-        lines = lines_left.union(lines_right)
-
-        return [(x in lines) for x in range(self.min_x, max(lines) + 1)]
-
-    def calculate_tiling(self, tiling):
-        h = [self.make_base_h_row(i) for i in range(self.min_y, self.max_y + 2)]
-        v = [self.make_base_v_row(i) for i in range(self.min_y, self.max_y + 1)]
-
-        for tile in tiling:
-            for sq_a in tile:
-                for sq_b in tile:
-                    a, b = sorted([sq_a, sq_b])
-                    ax, ay = a[0] - self.min_x, a[1] - self.min_y
-                    bx, by = b[0] - self.min_x, b[1] - self.min_y
-                    if (ay == by) and (ax + 1 == bx):
-                        v[ay][bx] = False
-                    if (ax == bx) and (ay + 1 == by):
-                        h[by][ax] = False
-        
-        return h, v
-
-    def format_tiling(self, tiling):
-        h, v = self.calculate_tiling(tiling)
-        return "\n".join(self.format_tiling_lines(h, v))
-
-    def display(self):
-        return self.format_tiling([self.squares])
-
 
 class DeletedRectangle(Irregular):
     def __init__(self, whole, deleted):
@@ -133,28 +98,34 @@ class DeletedRectangle(Irregular):
 
     def remove(self, square):
         if square not in self.squares:
-            raise Exception("Tried to remove a square %s which was not present in the board" % (square, ))
+            raise Exception(
+                "Tried to remove a square %s which was not present in the board"
+                % (square,)
+            )
         return DeletedRectangle(self.whole, self.deleted + [square])
-    
+
     def remove_all(self, squares):
         for square in squares:
             if square not in self.squares:
-                raise Exception("Tried to remove a square %s which was not present in the board" % (square, ))
+                raise Exception(
+                    "Tried to remove a square %s which was not present in the board"
+                    % (square,)
+                )
         return DeletedRectangle(self.whole, self.deleted + squares)
 
     @property
     def interior_deleted(self):
         return (
-            (x, y) for x, y in self.deleted 
-            if x > self.min_x and x < self.max_x
-            and y > self.min_y and y < self.max_y
+            (x, y)
+            for x, y in self.deleted
+            if x > self.min_x and x < self.max_x and y > self.min_y and y < self.max_y
         )
 
     def format_row_sides(self, row, row_y):
         line = list(super().format_row_sides(row))
         for x, y in self.interior_deleted:
             if y == row_y:
-                line[2 * x + 1] = 'X'
+                line[2 * x + 1] = "X"
         return "".join(line)
 
     def format_tiling_lines(self, h, v):
@@ -165,13 +136,17 @@ class DeletedRectangle(Irregular):
 
 
 class Rectangle(Shape):
+    _squares_cache = None
+
     def __init__(self, x, y):
         self.x = x
         self.y = y
 
     @property
     def squares(self):
-        return list(self.gen_squares())
+        if self._squares_cache is None:
+            self._squares_cache = list(self.gen_squares())
+        return self._squares_cache
 
     def gen_squares(self):
         for i in range(0, self.x):
@@ -206,7 +181,7 @@ class Rectangle(Shape):
                         pass
                     if (a[0] == b[0]) and (a[1] + 1 == b[1]):
                         h[b[1]][a[0]] = False
-        
+
         return h, v
 
     def format_tiling(self, tiling):
@@ -215,17 +190,22 @@ class Rectangle(Shape):
 
     def remove(self, square):
         if square not in self.squares:
-            raise Exception("Tried to remove a square %s which was not present in the board" % (square, ))
+            raise Exception(
+                "Tried to remove a square %s which was not present in the board"
+                % (square,)
+            )
         return DeletedRectangle(self, [square])
-    
+
     def remove_all(self, squares):
         for square in squares:
             if square not in self.squares:
-                raise Exception("Tried to remove a square %s which was not present in the board" % (square, ))
+                raise Exception(
+                    "Tried to remove a square %s which was not present in the board"
+                    % (square,)
+                )
         return DeletedRectangle(self, squares)
 
 
-
 class Chessboard(Rectangle):
     x = 8
     y = 8
